/***************************************************************************
    LibUFO - UI For OpenGL
    copyright         : (C) 2001-2005 by Johannes Schmidt
    email             : schmidtjf at users.sourceforge.net
                             -------------------

    file              : test/layouts.cpp
    begin             : Tue Dec 23 2003
    $Id: layouts.cpp,v 1.4 2005/02/14 10:45:03 schmidtjf Exp $
 ***************************************************************************/


#include <ufo/ufo.hpp>
#include <ufo/ux/ux.hpp>

// This example shows the use of more complex layouts

using namespace ufo;

void createContent(UWidget * content);

int
main(int argc, char ** argv) {
	// creates the toolkit object, initializes UFO
	UXToolkit tk;
	// loads the video driver and creates the event queue
	UXDisplay display;

	// creates a frame
	UXFrame * frame = display.createFrame();
	frame->setTitle("layout managers");
	frame->setBounds(0, 0, 640, 480);
	frame->setResizable(true);
	frame->setVisible(true);

	// Now you have a valid OpenGL context

	// user created widgets are placed in the content pane
	// create some widgets and add them to the content pane
	createContent(frame->getContentPane());

	// main event loop
	bool done = false;
	while (!done) {
		// pumps events from the system event queue
		display.pumpEvents();
		// dispatchEvents returns true when a quit event was processed.
		done = display.dispatchEvents();

		if (!done) {
			// repaint the frame
			frame->repaint();
			// give other processes a chance
			tk.sleep(1);
		}
	}

	return 0;
}

void
createContent(UWidget * content) {
	// this method adds some widgets to show the 4 basic layout manager
	//
	// the border layout:
	//  this layout can handle 5 childs.
	//   4 widgets for top, bottom, left and right
	//   1 centered widget
	//
	// the flow layout
	//  this layout handles an arbitrary amount of widgets, aligning
	//  them in on continuous flow (which may have line breaks).
	//
	// the box layout
	//  this layout aligns widgets in a vertical resp. horizontal box.
	//  Each widget gets the same width resp. height.
	//
	// the "NULL layout" or custom layout
	//  Actually not a layout manager. You have to set the location and
	//  the size yourself.

	// Structure of the layout managers in this example:
	//
	// We are using a border layout as top level layout.
	// We add a vertical box layout left.
	// In the middle, we show the flow layout.
	// On the right hand side, we add the NULL layout.
	// Below them, we add common Ok/Cancel widgets using
	//  a border layout an a box layout.
	//
	// Additionally, all containers get a line border to show their borders.

	// The content pane layout is the border layout
	content->setLayout(new UBorderLayout());

	// This is the box with vertical aligned widgets
	UWidget * verticalBox = new UWidget();
	verticalBox->setLayout(new UBoxLayout(UBoxLayout::YAxis));
	verticalBox->setBorder(LineBorder);

	// This is the box with the flow layout
	UWidget * flowLayoutBox = new UWidget();
	flowLayoutBox->setLayout(new UFlowLayout());
	flowLayoutBox->setBorder(LineBorder);

	// This is the NULL layout
	UWidget * nullLayoutBox = new UWidget();
	nullLayoutBox->setLayout(NULL);
	// We have to set a preferred size. Otherwise, the size would be 0.
	nullLayoutBox->setPreferredSize(UDimension(200, 200));
	nullLayoutBox->setBorder(LineBorder);

	// This is the Ok/Cancel code
	UWidget * rightBox = new UWidget();
	rightBox->setLayout(new UBoxLayout(UBoxLayout::XAxis));
	rightBox->setBorder(LineBorder);

	UWidget * rightBoxParent = new UWidget();
	rightBoxParent->setLayout(new UBorderLayout());
	rightBoxParent->add(rightBox, UBorderLayout::East);
	rightBoxParent->setBorder(LineBorder);

	// Add the three boxes to the content pane

	// verticalBox in the west
	content->add(verticalBox, UBorderLayout::West);

	// flowLayoutBox in the center
	content->add(flowLayoutBox, UBorderLayout::Center);

	// nullLayoutBox in the east
	content->add(nullLayoutBox, UBorderLayout::East);

	// rightBox in the south
	content->add(rightBoxParent, UBorderLayout::South);


	//
	// now add some content
	//
	verticalBox->add(new ULabel("verticalBox 1"));
	verticalBox->add(new ULabel("verticalBox 2"));
	verticalBox->add(new ULabel("verticalBox 3"));

	flowLayoutBox->add(new ULabel("flowLayoutBox 1"));
	flowLayoutBox->add(new ULabel("flowLayoutBox 2"));
	flowLayoutBox->add(new ULabel("flowLayoutBox 3"));

	// null layout
	ULabel * label = new ULabel("custom layout label");
	// To get the preferred size, the label _has_ to be added to a valid
	// widget visible on screen
	nullLayoutBox->add(label);
	label->setLocation(10, 10);
	label->setSize(label->getPreferredSize());

	ULabel * label2 = new ULabel("custom layout label2");
	nullLayoutBox->add(label2);
	label2->setLocation(10, 60);
	label2->setSize(label2->getPreferredSize());

	// note: these buttons do nothing
	rightBox->add(new UButton("Ok"));
	rightBox->add(new UButton("Cancel"));
}
