/***************************************************************************
    LibUFO - UI For OpenGL
    copyright         : (C) 2001-2005 by Johannes Schmidt
    email             : schmidtjf at users.sourceforge.net
                             -------------------

    file              : test/widget3d.cpp
    begin             : Sat Jan 17 2004
    $Id: widget3d.cpp,v 1.5 2005/02/14 10:45:03 schmidtjf Exp $
 ***************************************************************************/


#include <ufo/ufo.hpp>
#include <ufo/ux/ux.hpp>

#include <ufo/ufo_gl.hpp>

// This example shows how to use custom OpenGL code in your apps.
// Another way is shown in gltest.cpp (without widgets).

using namespace ufo;

void createContent(UWidget * content);

// 3d widget class definition

/** This widget overrides UWidget::paintWidget() to do custom drawing */
class Widget3D : public UWidget {
public:
	Widget3D();
protected: // Overrides UWidget
	/** Overrides the default paintWidget implementation and
	  * calls our draw3D method.
	  */
	void paintWidget(UGraphics * g);

protected: // Protected methods
	/** Inits the projection matrix. */
	void initProjection(UGraphics * g);
	/** Resets the projection matrix to the old one. */
	void resetProjection(UGraphics * g);

	/** Our 3d code. */
	void draw3D();

private:
	/** Caches the viewport bounds. */
	GLint vPort[4];
};


int
main(int argc, char ** argv) {
	// creates the toolkit object, initializes UFO
	UXToolkit tk;
	// loads the video driver and creates the event queue
	UXDisplay display;

	// creates a frame
	UXFrame * frame = display.createFrame();
	frame->setTitle("3d widgets");
	frame->setBounds(0, 0, 640, 480);
	frame->setResizable(true);
	frame->setVisible(true);

	// Now you have a valid OpenGL context

	// user created widgets are placed in the content pane
	// create some widgets and add them to the content pane
	createContent(frame->getContentPane());

	// main event loop
	bool done = false;
	while (!done) {
		// pumps events from the system event queue
		display.pumpEvents();
		// dispatchEvents returns true when a quit event was processed.
		done = display.dispatchEvents();

		if (!done) {
			// repaint the frame
			frame->repaint();
			// give other processes a chance
			tk.sleep(1);
		}
	}

	return 0;
}

void
createContent(UWidget * content) {
	// we are using the flow layout as layout manager
	content->setLayout(new UFlowLayout());

	// add a hello world label (static text)
	content->add(new ULabel("Hello, World!"));


	// add our custom opengl widget
	content->add(new Widget3D());
}

//
// Widget3D implementation
//

// we need the opengl headers
#include <GL/gl.h>

// copied from testgl.c of test directory of the SDL1.2.3 sources
// http://www.libsdl.org
static const float color[8][3]= {
	{ 1.0,  1.0,  0.0},
	{ 1.0,  0.0,  0.0},
	{ 0.0,  0.0,  0.0},
	{ 0.0,  1.0,  0.0},
	{ 0.0,  1.0,  1.0},
	{ 1.0,  1.0,  1.0},
	{ 1.0,  0.0,  1.0},
	{ 0.0,  0.0,  1.0}
};
static const float cube[8][3]= {
	{ 0.5,  0.5, -0.5},
	{ 0.5, -0.5, -0.5},
	{-0.5, -0.5, -0.5},
	{-0.5,  0.5, -0.5},
	{-0.5,  0.5,  0.5},
	{ 0.5,  0.5,  0.5},
	{ 0.5, -0.5,  0.5},
	{-0.5, -0.5,  0.5}
};


Widget3D::Widget3D() {
	// Set the preferred size
	// This is necessarry as we use a layout manager and
	// UWidget has a default preferred size of 0.
	setPreferredSize(UDimension(200, 200));
}


void
Widget3D::paintWidget(UGraphics * g) {
	// enable some attributes we want to use
	glClear(GL_DEPTH_BUFFER_BIT);
	glEnable(GL_DEPTH_TEST);

	glShadeModel(GL_SMOOTH);

	// init our projection and model view matrix
	initProjection(g);

	// our 3d code
	draw3D();

	// reset matrices
	resetProjection(g);

	// reset attributes
	glDisable(GL_DEPTH_TEST);
	glShadeModel(GL_FLAT);

	// schedule a repaint for this widget
	repaint();
}


void
Widget3D::initProjection(UGraphics * g) {
	// At first, we are replacing the old viewport with
	// a viewport matching the bounds of our widgets.

	// Then we change the projection and model view matrix


	// save the old viewport coordinates (for resetting)
	glGetIntegerv(GL_VIEWPORT, vPort);

	// converts the root bounds (bounds relative to the topmost root pane)
	// of this 3d widget into OpenGL coordinates
	URectangle openglRect = g->mapToDevice(getRootBounds());

	// set the new viewport
	glViewport(openglRect.x, openglRect.y, openglRect.w, openglRect.h);

	// set up projection and model view matrix
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();

	glOrtho(-2.0, 2.0, -2.0, 2.0, -20.0, 20.0);

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();

	// rotate (for rotating cube)
	static float rot = 0.0f;
	glRotatef(rot, 1.0, 1.0, 1.0);
	rot += 5.0f;

	if (rot >360.0f) {
		rot = 0.0f;
	}
}

void
Widget3D::resetProjection(UGraphics * g) {
	// reset viewport with the old values
	glViewport(vPort[0], vPort[1], vPort[2], vPort[3]);

	// reset projection and model view matrix
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
}

void
Widget3D::draw3D() {
	// draw a colored cube
	glBegin(GL_QUADS);

	glColor3fv(color[0]);
	glVertex3fv(cube[0]);
	glColor3fv(color[1]);
	glVertex3fv(cube[1]);
	glColor3fv(color[2]);
	glVertex3fv(cube[2]);
	glColor3fv(color[3]);
	glVertex3fv(cube[3]);

	glColor3fv(color[3]);
	glVertex3fv(cube[3]);
	glColor3fv(color[4]);
	glVertex3fv(cube[4]);
	glColor3fv(color[7]);
	glVertex3fv(cube[7]);
	glColor3fv(color[2]);
	glVertex3fv(cube[2]);

	glColor3fv(color[0]);
	glVertex3fv(cube[0]);
	glColor3fv(color[5]);
	glVertex3fv(cube[5]);
	glColor3fv(color[6]);
	glVertex3fv(cube[6]);
	glColor3fv(color[1]);
	glVertex3fv(cube[1]);

	glColor3fv(color[5]);
	glVertex3fv(cube[5]);
	glColor3fv(color[4]);
	glVertex3fv(cube[4]);
	glColor3fv(color[7]);
	glVertex3fv(cube[7]);
	glColor3fv(color[6]);
	glVertex3fv(cube[6]);

	glColor3fv(color[5]);
	glVertex3fv(cube[5]);
	glColor3fv(color[0]);
	glVertex3fv(cube[0]);
	glColor3fv(color[3]);
	glVertex3fv(cube[3]);
	glColor3fv(color[4]);
	glVertex3fv(cube[4]);

	glColor3fv(color[6]);
	glVertex3fv(cube[6]);
	glColor3fv(color[1]);
	glVertex3fv(cube[1]);
	glColor3fv(color[2]);
	glVertex3fv(cube[2]);
	glColor3fv(color[7]);
	glVertex3fv(cube[7]);

	glEnd();
}
